@@ -1,16 +1,31 @@
-/* @(#)pch.c	1.16 11/05/22 2011 J. Schilling */
+/* @(#)pch.c	1.29 15/06/03 2011-2015 J. Schilling */
 #include <schily/mconfig.h>
 #ifndef lint
 static	UConst char sccsid[] =
-	"@(#)pch.c	1.16 11/05/22 2011 J. Schilling";
+	"@(#)pch.c	1.29 15/06/03 2011-2015 J. Schilling";
 #endif
 /*
  *	Copyright (c) 1986-1988 Larry Wall
  *	Copyright (c) 1990 Wayne Davison
- *	Copyright (c) 2011 J. Schilling
+ *	Copyright (c) 2011-2015 J. Schilling
  *
- *	This program may be copied as long as you don't try to make any
- *	money off of it, or pretend that you wrote it.
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following condition is met:
+ *
+ * 1. Redistributions of source code must retain the above copyright notice,
+ * this condition and the following disclaimer.
+ * 
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND ANY
+ * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
+ * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+ * DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE FOR
+ * ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
+ * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+ * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
  */
 
 #define	EXT	extern
@@ -27,13 +42,13 @@ static off_t	p_filesize;		/* size of the patch file */
 static LINENUM p_first;			/* 1st line number */
 static LINENUM p_newfirst;		/* 1st line number of replacement */
 static LINENUM p_ptrn_lines;		/* # lines in pattern */
-static LINENUM p_repl_lines;		/* # lines in replacement text */
+EXT    LINENUM p_repl_lines;		/* # lines in replacement text */
 static LINENUM p_end = -1;		/* last line in hunk */
 static LINENUM p_max;			/* max allowed value of p_end */
 static LINENUM p_context = 3;		/* # of context lines */
 static LINENUM p_input_line = 0;	/* current line # from patch file */
 static char **p_line = Null(char **);	/* the text of the hunk */
-static short *p_len = Null(short *);	/* length of each line */
+static size_t *p_len = Null(size_t *);	/* length of each line */
 static char *p_char = Nullch;		/* +, -, and ! */
 static LINENUM hunkmax = INITHUNKMAX;	/* size of above arrays to begin with */
 static size_t p_indent;			/* indent to patch */
@@ -50,7 +65,8 @@ static	int	intuit_diff_type __PR((void));
 static	void	next_intuit_at __PR((LINENUM file_pos, LINENUM file_line));
 static	void	skip_to __PR((off_t file_pos, off_t file_line));
 static	void	malformed __PR((void));
-static	char	*pgets __PR((char *bf, int sz, FILE *fp));
+static	void	ovstrcpy __PR((char *p2, char *p1));
+static	ssize_t	pgets __PR((char **bfp, size_t *szp, FILE *fp));
 
 /* Prepare to look for the next patch in the patch file. */
 
@@ -75,15 +91,15 @@ open_patch_file(filename)
 	if (filename == Nullch || !*filename || strEQ(filename, "-")) {
 		pfp = fopen(TMPPATNAME, "w");
 		if (pfp == Nullfp)
-			fatal(_("patch: can't create %s.\n"), TMPPATNAME);
-		while (fgets(buf, sizeof (buf), stdin) != Nullch)
+			pfatal(_("can't create %s.\n"), TMPPATNAME);
+		while (fgetaline(stdin, &buf, &bufsize) > 0)
 			fputs(buf, pfp);
 		Fclose(pfp);
 		filename = TMPPATNAME;
 	}
 	pfp = fopen(filename, "r");
 	if (pfp == Nullfp)
-		fatal(_("patch file %s not found\n"), filename);
+		pfatal(_("patch file %s not found\n"), filename);
 	Fstat(fileno(pfp), &file_stat);
 	p_filesize = file_stat.st_size;
 	next_intuit_at((LINENUM)0, (LINENUM)1);	/* start at the beginning */
@@ -97,8 +113,8 @@ set_hunkmax()
 {
 	if (p_line == Null(char **))
 		p_line = (char **) malloc(hunkmax * sizeof (char *));
-	if (p_len == Null(short *))
-		p_len  = (short *) malloc(hunkmax * sizeof (short));
+	if (p_len == Null(size_t *))
+		p_len  = (size_t *) malloc(hunkmax * sizeof (size_t));
 	if (p_char == Nullch)
 		p_char = (char *)  malloc(hunkmax * sizeof (char));
 }
@@ -108,6 +124,10 @@ set_hunkmax()
 static void
 grow_hunkmax()
 {
+	char	**op_line = p_line;
+	size_t	*op_len = p_len;
+	char	*op_char = p_char;
+
 	hunkmax *= 2;
 	/*
 	 * Note that on most systems, only the p_line array ever gets fresh
@@ -116,18 +136,27 @@ grow_hunkmax()
 	 * doesn't matter.
 	 */
 	assert(p_line != Null(char **) &&
-	    p_len != Null(short *) &&
+	    p_len != Null(size_t *) &&
 	    p_char != Nullch);
 	p_line = (char **) realloc((char *)p_line, hunkmax * sizeof (char *));
-	p_len  = (short *) realloc((char *)p_len,  hunkmax * sizeof (short));
+	p_len  = (size_t *) realloc((char *)p_len,  hunkmax * sizeof (size_t));
 	p_char = (char *)  realloc((char *)p_char, hunkmax * sizeof (char));
 	if (p_line != Null(char **) &&
-	    p_len != Null(short *) &&
+	    p_len != Null(size_t *) &&
 	    p_char != Nullch) {
 		return;
 	}
+	if (p_line == Null(char **))
+		p_line = op_line;
+	if (p_len == Null(size_t *))
+		p_len = op_len;
+	if (p_char == Null(char *))
+		p_char = op_char;
+	hunkmax /= 2;
+
 	if (!using_plan_a)
 		fatal(_("patch: out of memory (grow_hunkmax)\n"));
+
 	out_of_mem = TRUE;	/* whatever is null will be allocated again */
 				/* from within plan_a(), of all places */
 }
@@ -235,7 +264,7 @@ intuit_diff_type()
 		this_line = ftell(pfp);
 		indent = 0;
 		p_input_line++;
-		if (fgets(buf, sizeof (buf), pfp) == Nullch) {
+		if (fgetaline(pfp, &buf, &bufsize) <= 0) {
 			if (first_command_line >= 0L) {
 				/* nothing but deletes!? */
 				p_start = first_command_line;
@@ -255,16 +284,30 @@ intuit_diff_type()
 			else
 				indent++;
 		}
-		for (t = s; isdigit(*t) || *t == ','; t++) {
+		for (t = s; isdigit(UCH *t) || *t == ','; t++) {
 			;
 			/* LINTED */
 		}
-		this_is_a_command = (isdigit(*s) &&
+		this_is_a_command = (isdigit(UCH *s) &&
 		    (*t == 'd' || *t == 'c' || *t == 'a'));
 		if (first_command_line < 0L && this_is_a_command) {
-		first_command_line = this_line;
-		fcl_line = p_input_line;
-		p_indent = indent;		/* assume this for now */
+			first_command_line = this_line;
+			fcl_line = p_input_line;
+			p_indent = indent;	/* assume this for now */
+
+			if (t[1] == '\n') {
+				p_start = first_command_line;
+				p_sline = fcl_line;
+				retval = ED_DIFF;
+				goto scan_exit;
+			}
+		}
+		if (first_command_line < 0L && isdigit(UCH *s) &&
+		    (*t == 'i' || *t == 's')) {
+			p_start = first_command_line;
+			p_sline = fcl_line;
+			retval = ED_DIFF;
+			goto scan_exit;
 		}
 		if (!stars_last_line && strnEQ(s, "*** ", 4)) {
 			oldtmp = savestr(s+4);
@@ -276,12 +319,12 @@ intuit_diff_type()
 			indtmp = savestr(s+6);
 		} else if (wall_plus &&
 			    strnEQ(s, "Prereq:", 7)) {
-			for (t = s+7; isspace(*t); t++) {
+			for (t = s+7; isspace(UCH *t); t++) {
 				;
 				/* LINTED */
 			}
 			revision = savestr(t);
-			for (t = revision; *t && !isspace(*t); t++) {
+			for (t = revision; *t && !isspace(UCH *t); t++) {
 				;
 				/* LINTED */
 			}
@@ -302,19 +345,27 @@ intuit_diff_type()
 		}
 		if ((!diff_type || diff_type == UNI_DIFF) &&
 		    strnEQ(s, "@@ -", 4)) {
-			if (!atol(s+3))
+			if (!atolnum(s+3))
 				ok_to_create_file = TRUE;
 			p_indent = indent;
 			p_start = this_line;
 			p_sline = p_input_line;
 			retval = UNI_DIFF;
+			/*
+			 * "--- " was scanned as "newname" and
+			 * "+++ " was scanned as "oldname".
+			 * Swap names for correct POSIX name selcetion.
+			 */
+			t = newtmp;
+			newtmp = oldtmp;
+			oldtmp = t;
 			goto scan_exit;
 		}
 		stars_this_line = strnEQ(s, "********", 8);
 		if ((!diff_type || diff_type == CONTEXT_DIFF) &&
 		    stars_last_line &&
 		    strnEQ(s, "*** ", 4)) {
-			if (!atol(s+4))
+			if (!atolnum(s+4))
 				ok_to_create_file = TRUE;
 			/* if this is a new context diff the character just */
 			/* before the newline is a '*'. */
@@ -352,7 +403,14 @@ intuit_diff_type()
 			newname = fetchname(newtmp, strippath,
 						ok_to_create_file,
 						&is_null_time[1]);
-		if (oldname && newname) {
+
+		if (do_wall && oldname && newname) {
+			/*
+			 * Old Larry Wall algorithm from patch-2.0.
+			 * This heuristic works only if the old name is
+			 * something like "file.orig" and the new name is just
+			 * "file". For "original" vs. "changed", it fails.
+			 */
 			if (strlen(oldname) < strlen(newname))
 				filearg[0] = savestr(oldname);
 			else
@@ -369,16 +427,23 @@ intuit_diff_type()
 		free(bestguess);
 		bestguess = Nullch;
 	}
-	if (filearg[0] != Nullch)
+	if (filearg[0] != Nullch) {
 		bestguess = savestr(filearg[0]);
-	else if (indtmp != Nullch)
+	} else if (indtmp != Nullch) {
 		bestguess = fetchname(indtmp, strippath, TRUE, NULL);
-	else {
+	} else {
 		if (oldtmp != Nullch)
 			oldname = fetchname(oldtmp, strippath, TRUE, NULL);
 		if (newtmp != Nullch)
 			newname = fetchname(newtmp, strippath, TRUE, NULL);
-		if (oldname && newname) {
+
+		if (do_wall && oldname && newname) {
+			/*
+			 * Old Larry Wall algorithm from patch-2.0.
+			 * This heuristic works only if the old name is
+			 * something like "file.orig" and the new name is just
+			 * "file". For "original" vs. "changed", it fails.
+			 */
 			if (strlen(oldname) < strlen(newname))
 				bestguess = savestr(oldname);
 			else
@@ -422,7 +487,7 @@ skip_to(file_pos, file_line)
 	off_t file_pos;
 	off_t file_line;
 {
-	char *ret;
+	ssize_t	ret;
 
 	assert(p_base <= file_pos);
 	if (verbose && p_base < file_pos) {
@@ -430,8 +495,8 @@ skip_to(file_pos, file_line)
 		say(
 _("The text leading up to this was:\n--------------------------\n"));
 		while (ftell(pfp) < file_pos) {
-			ret = fgets(buf, sizeof (buf), pfp);
-			assert(ret != Nullch);
+			ret = fgetaline(pfp, &buf, &bufsize);
+			assert(ret > 0);
 			say("|%s", buf);
 		}
 		say("--------------------------\n");
@@ -455,7 +520,7 @@ bool
 another_hunk()
 {
 	char	*s;
-	char	*ret;
+	ssize_t	ret;
 	LINENUM	context = 0;
 
 	while (p_end >= 0) {
@@ -488,9 +553,9 @@ another_hunk()
 
 		fillsrc = filldst = repl_patch_line = 0; /* Make gcc quiet */
 
-		ret = pgets(buf, sizeof (buf), pfp);
+		ret = pgets(&buf, &bufsize, pfp);
 		p_input_line++;
-		if (ret == Nullch || strnNE(buf, "********", 8)) {
+		if (ret <= 0 || strnNE(buf, "********", 8)) {
 			next_intuit_at(line_beginning, p_input_line);
 			return (FALSE);
 		}
@@ -498,9 +563,9 @@ another_hunk()
 		p_hunk_beg = p_input_line + 1;
 		while (p_end < p_max) {
 			line_beginning = ftell(pfp);
-			ret = pgets(buf, sizeof (buf), pfp);
+			ret = pgets(&buf, &bufsize, pfp);
 			p_input_line++;
-			if (ret == Nullch) {
+			if (ret <= 0) {
 				if (p_max - p_end < 4)
 					Strcpy(buf, "  \n");  /* assume blank lines got chopped */
 				else {
@@ -550,7 +615,7 @@ _("Unexpected end of hunk at line %lld.\n"),
 					p_end--;
 					return (FALSE);
 				}
-				for (s = buf; *s && !isdigit(*s); s++) {
+				for (s = buf; *s && !isdigit(UCH *s); s++) {
 					;
 					/* LINTED */
 				}
@@ -558,17 +623,17 @@ _("Unexpected end of hunk at line %lld.\n"),
 					malformed();
 				if (strnEQ(s, "0,0", 3))
 					strcpy(s, s+2);
-				p_first = (LINENUM) atol(s);
-				while (isdigit(*s))
+				p_first = atolnum(s);
+				while (isdigit(UCH *s))
 					s++;
 				if (*s == ',') {
-					for (; *s && !isdigit(*s); s++) {
+					for (; *s && !isdigit(UCH *s); s++) {
 						;
 						/* LINTED */
 					}
 					if (!*s)
 						malformed();
-					p_ptrn_lines = ((LINENUM)atol(s)) -
+					p_ptrn_lines = atolnum(s) -
 								p_first + 1;
 				} else if (p_first) {
 					p_ptrn_lines = 1;
@@ -577,9 +642,13 @@ _("Unexpected end of hunk at line %lld.\n"),
 					p_first = 1;
 				}
 				p_max = p_ptrn_lines + 6;	/* we need this much at least */
-				while (p_max >= hunkmax)
+				while (p_max >= hunkmax && !out_of_mem)
 					grow_hunkmax();
 				p_max = hunkmax;
+				if (out_of_mem) {
+					p_end = -1;
+					return (FALSE);
+				}
 				break;
 			case '-':
 				if (buf[1] == '-') {
@@ -623,23 +692,26 @@ _("%s \"---\" at line %lld--check line numbers at line %lld.\n"),
 						return (FALSE);
 					}
 					p_char[p_end] = '=';
-					for (s = buf; *s && !isdigit(*s); s++) {
+					for (s = buf;
+					    *s && !isdigit(UCH *s); s++) {
 						;
 						/* LINTED */
 					}
 					if (!*s)
 						malformed();
-					p_newfirst = (LINENUM) atol(s);
-					while (isdigit(*s))
+					p_newfirst = atolnum(s);
+					while (isdigit(UCH *s))
 						s++;
 					if (*s == ',') {
-						for (; *s && !isdigit(*s); s++) {
+						for (;
+						    *s && !isdigit(UCH *s);
+						    s++) {
 							;
 							/* LINTED */
 						}
 						if (!*s)
 							malformed();
-						p_repl_lines = ((LINENUM)atol(s)) - p_newfirst + 1;
+						p_repl_lines = atolnum(s) - p_newfirst + 1;
 					} else if (p_newfirst) {
 						p_repl_lines = 1;
 					} else {
@@ -647,13 +719,12 @@ _("%s \"---\" at line %lld--check line numbers at line %lld.\n"),
 						p_newfirst = 1;
 					}
 					p_max = p_repl_lines + p_end;
-					if (p_max > MAXHUNKSIZE)
-						fatal(
-_("Hunk too large (%lld lines) at line %lld: %s"),
-						    (Llong)p_max,
-						    (Llong)p_input_line, buf);
-					while (p_max >= hunkmax)
+					while (p_max >= hunkmax && !out_of_mem)
 						grow_hunkmax();
+					if (out_of_mem) {
+						p_end = -1;
+						return (FALSE);
+					}
 					if (p_repl_lines != ptrn_copiable &&
 					    (p_context != 0 || p_repl_lines != 1))
 						repl_could_be_missing = FALSE;
@@ -666,7 +737,8 @@ _("Hunk too large (%lld lines) at line %lld: %s"),
 			change_line:
 				if (buf[1] == '\n' && canonicalize)
 					strcpy(buf+1, " \n");
-				if (!isspace(buf[1]) && buf[1] != '>' && buf[1] != '<' &&
+				if (!isspace(UCH buf[1]) &&
+				    buf[1] != '>' && buf[1] != '<' &&
 				    repl_beginning && repl_could_be_missing) {
 					repl_missing = TRUE;
 					goto hunk_done;
@@ -703,7 +775,7 @@ _("Hunk too large (%lld lines) at line %lld: %s"),
 				}
 				break;
 			case ' ':
-				if (!isspace(buf[1]) &&
+				if (!isspace(UCH buf[1]) &&
 				    repl_beginning && repl_could_be_missing) {
 					repl_missing = TRUE;
 					goto hunk_done;
@@ -817,21 +889,21 @@ _("Replacement text or line numbers mangled in hunk at line %lld\n"),
 		LINENUM filldst;		/* index of new lines */
 		char ch;
 
-		ret = pgets(buf, sizeof (buf), pfp);
+		ret = pgets(&buf, &bufsize, pfp);
 		p_input_line++;
-		if (ret == Nullch || strnNE(buf, "@@ -", 4)) {
+		if (ret <= 0 || strnNE(buf, "@@ -", 4)) {
 			next_intuit_at(line_beginning, p_input_line);
 			return (FALSE);
 		}
 		s = buf+4;
 		if (!*s)
 			malformed();
-		p_first = (LINENUM) atol(s);
-		while (isdigit(*s))
+		p_first = atolnum(s);
+		while (isdigit(UCH *s))
 			s++;
 		if (*s == ',') {
-			p_ptrn_lines = (LINENUM) atol(++s);
-			while (isdigit(*s))
+			p_ptrn_lines = atolnum(++s);
+			while (isdigit(UCH *s))
 				s++;
 		} else {
 			p_ptrn_lines = 1;
@@ -840,12 +912,12 @@ _("Replacement text or line numbers mangled in hunk at line %lld\n"),
 			s++;
 		if (*s != '+' || !*++s)
 			malformed();
-		p_newfirst = (LINENUM) atol(s);
-		while (isdigit(*s))
+		p_newfirst = atolnum(s);
+		while (isdigit(UCH *s))
 			s++;
 		if (*s == ',') {
-			p_repl_lines = (LINENUM) atol(++s);
-			while (isdigit(*s)) s++;
+			p_repl_lines = atolnum(++s);
+			while (isdigit(UCH *s)) s++;
 		} else {
 			p_repl_lines = 1;
 		}
@@ -856,8 +928,12 @@ _("Replacement text or line numbers mangled in hunk at line %lld\n"),
 		if (!p_ptrn_lines)
 			p_first++;	/* do append rather than insert */
 		p_max = p_ptrn_lines + p_repl_lines + 1;
-		while (p_max >= hunkmax)
+		while (p_max >= hunkmax && !out_of_mem)
 			grow_hunkmax();
+		if (out_of_mem) {
+			p_end = -1;
+			return (FALSE);
+		}
 		fillsrc = 1;
 		filldst = fillsrc + p_ptrn_lines;
 		p_end = filldst + p_repl_lines;
@@ -882,9 +958,9 @@ _("Replacement text or line numbers mangled in hunk at line %lld\n"),
 		p_hunk_beg = p_input_line + 1;
 		while (fillsrc <= p_ptrn_lines || filldst <= p_end) {
 			line_beginning = ftell(pfp);
-			ret = pgets(buf, sizeof (buf), pfp);
+			ret = pgets(&buf, &bufsize, pfp);
 			p_input_line++;
-			if (ret == Nullch) {
+			if (ret <= 0) {
 				if (p_max - filldst < 3) {
 					Strcpy(buf, " \n");  /* assume blank lines got chopped */
 				} else {
@@ -970,20 +1046,20 @@ _("Unexpected end of file in patch.\n"));
 		off_t	line_beginning = ftell(pfp);
 
 		p_context = 0;
-		ret = pgets(buf, sizeof (buf), pfp);
+		ret = pgets(&buf, &bufsize, pfp);
 		p_input_line++;
-		if (ret == Nullch || !isdigit(*buf)) {
+		if (ret <= 0 || !isdigit(UCH *buf)) {
 			next_intuit_at(line_beginning, p_input_line);
 			return (FALSE);
 		}
-		p_first = (LINENUM)atol(buf);
-		for (s = buf; isdigit(*s); s++) {
+		p_first = atolnum(buf);
+		for (s = buf; isdigit(UCH *s); s++) {
 			;
 			/* LINTED */
 		}
 		if (*s == ',') {
-			p_ptrn_lines = (LINENUM)atol(++s) - p_first + 1;
-			while (isdigit(*s))
+			p_ptrn_lines = atolnum(++s) - p_first + 1;
+			while (isdigit(UCH *s))
 				s++;
 		} else {
 			p_ptrn_lines = (*s != 'a');
@@ -991,23 +1067,24 @@ _("Unexpected end of file in patch.\n"));
 		hunk_type = *s;
 		if (hunk_type == 'a')
 			p_first++;	/* do append rather than insert */
-		min = (LINENUM)atol(++s);
-		for (; isdigit(*s); s++) {
+		min = atolnum(++s);
+		for (; isdigit(UCH *s); s++) {
 			;
 			/* LINTED */
 		}
 		if (*s == ',')
-			max = (LINENUM)atol(++s);
+			max = atolnum(++s);
 		else
 			max = min;
 		if (hunk_type == 'd')
 			min++;
 		p_end = p_ptrn_lines + 1 + max - min + 1;
-		if (p_end > MAXHUNKSIZE)
-			fatal(_("Hunk too large (%lld lines) at line %lld: %s"),
-			    (Llong)p_end, (Llong)p_input_line, buf);
-		while (p_end >= hunkmax)
+		while (p_end >= hunkmax && !out_of_mem)
 			grow_hunkmax();
+		if (out_of_mem) {
+			p_end = -1;
+			return (FALSE);
+		}
 		p_newfirst = min;
 		p_repl_lines = max - min + 1;
 		Sprintf(buf, "*** %lld,%lld\n",
@@ -1019,9 +1096,9 @@ _("Unexpected end of file in patch.\n"));
 		}
 		p_char[0] = '*';
 		for (i = 1; i <= p_ptrn_lines; i++) {
-			ret = pgets(buf, sizeof (buf), pfp);
+			ret = pgets(&buf, &bufsize, pfp);
 			p_input_line++;
-			if (ret == Nullch)
+			if (ret <= 0)
 				fatal(
 _("Unexpected end of file in patch at line %lld.\n"),
 				    (Llong)p_input_line);
@@ -1037,9 +1114,9 @@ _("Unexpected end of file in patch at line %lld.\n"),
 			p_char[i] = '-';
 		}
 		if (hunk_type == 'c') {
-			ret = pgets(buf, sizeof (buf), pfp);
+			ret = pgets(&buf, &bufsize, pfp);
 			p_input_line++;
-			if (ret == Nullch)
+			if (ret <= 0)
 				fatal(
 _("Unexpected end of file in patch at line %lld.\n"),
 				    (Llong)p_input_line);
@@ -1055,9 +1132,9 @@ _("Unexpected end of file in patch at line %lld.\n"),
 		}
 		p_char[i] = '=';
 		for (i++; i <= p_end; i++) {
-			ret = pgets(buf, sizeof (buf), pfp);
+			ret = pgets(&buf, &bufsize, pfp);
 			p_input_line++;
-			if (ret == Nullch)
+			if (ret <= 0)
 				fatal(
 _("Unexpected end of file in patch at line %lld.\n"),
 				    (Llong)p_input_line);
@@ -1097,20 +1174,32 @@ _("Unexpected end of file in patch at line %lld.\n"),
 	return (TRUE);
 }
 
+/*
+ * A strcpy() that works with overlapping buffers
+ */
+static void
+ovstrcpy(p2, p1)
+	register char	*p2;
+	register char	*p1;
+{
+	while ((*p2++ = *p1++) != '\0')
+		;
+}
+
 /* Input a line from the patch file, worrying about indentation. */
 
-static char *
-pgets(bf, sz, fp)
-	char	*bf;
-	int	sz;
+static ssize_t
+pgets(bfp, szp, fp)
+	char	**bfp;
+	size_t	*szp;
 	FILE	*fp;
 {
-	char *ret = fgets(bf, sz, fp);
+	ssize_t	ret = fgetaline(fp, bfp, szp);
 	char *s;
 	size_t indent = 0;
 
-	if (p_indent && ret != Nullch) {
-		for (s = buf;
+	if (p_indent && ret > 0) {
+		for (s = *bfp;
 		    indent < p_indent && (*s == ' ' || *s == '\t' || *s == 'X');
 		    s++) {
 			if (*s == '\t')
@@ -1118,8 +1207,8 @@ pgets(bf, sz, fp)
 			else
 				indent++;
 		}
-		if (buf != s)
-			Strcpy(buf, s);
+		if (*bfp != s)
+			ovstrcpy(*bfp, s);
 	}
 	return (ret);
 }
@@ -1130,7 +1219,7 @@ bool
 pch_swap()
 {
 	char **tp_line;		/* the text of the hunk */
-	short *tp_len;		/* length of each line */
+	size_t *tp_len;		/* length of each line */
 	char *tp_char;		/* +, -, and ! */
 	LINENUM i;
 	LINENUM n;
@@ -1147,19 +1236,19 @@ pch_swap()
 	tp_len = p_len;
 	tp_char = p_char;
 	p_line = Null(char **);	/* force set_hunkmax to allocate again */
-	p_len = Null(short *);
+	p_len = Null(size_t *);
 	p_char = Nullch;
 	set_hunkmax();
 	if (p_line == Null(char **) ||
-	    p_len == Null(short *) ||
+	    p_len == Null(size_t *) ||
 	    p_char == Nullch) {
-		if (p_line == Null(char **))
+		if (p_line != Null(char **))
 			free((char *)p_line);
 		p_line = tp_line;
-		if (p_len == Null(short *))
+		if (p_len != Null(size_t *))
 			free((char *)p_len);
 		p_len = tp_len;
-		if (p_char == Nullch)
+		if (p_char != Nullch)
 			free((char *)p_char);
 		p_char = tp_char;
 		return (FALSE);		/* not enough memory to swap hunk! */
@@ -1218,11 +1307,11 @@ pch_swap()
 	i = p_ptrn_lines;
 	p_ptrn_lines = p_repl_lines;
 	p_repl_lines = i;
-	if (tp_line == Null(char **))
+	if (tp_line != Null(char **))
 		free((char *)tp_line);
-	if (tp_len == Null(short *))
+	if (tp_len != Null(size_t *))
 		free((char *)tp_len);
-	if (tp_char == Nullch)
+	if (tp_char != Nullch)
 		free((char *)tp_char);
 	return (TRUE);
 }
@@ -1277,7 +1366,7 @@ pch_context()
 
 /* Return the length of a particular patch line. */
 
-short
+size_t
 pch_line_len(line)
 LINENUM line;
 {
@@ -1331,23 +1420,22 @@ do_ed_script()
 	}
 	for (;;) {
 		beginning_of_this_line = ftell(pfp);
-		if (pgets(buf, sizeof (buf), pfp) == Nullch) {
+		if (pgets(&buf, &bufsize, pfp) <= 0) {
 			next_intuit_at(beginning_of_this_line, p_input_line);
 			break;
 		}
 		p_input_line++;
-		for (t = buf; isdigit(*t) || *t == ','; t++) {
+		for (t = buf; isdigit(UCH *t) || *t == ','; t++) {
 			;
 			/* LINTED */
 		}
-		this_line_is_command = (isdigit(*buf) &&
+		this_line_is_command = (isdigit(UCH *buf) &&
 		    (*t == 'd' || *t == 'c' || *t == 'a'));
 		if (this_line_is_command) {
 			if (!skip_rest_of_patch)
 				fputs(buf, pipefp);
 			if (*t != 'd') {
-				while (pgets(buf, sizeof (buf), pfp) !=
-				    Nullch) {
+				while (pgets(&buf, &bufsize, pfp) > 0) {
 					p_input_line++;
 					if (!skip_rest_of_patch)
 						fputs(buf, pipefp);
@@ -1375,3 +1463,66 @@ do_ed_script()
 	}
 	set_signals(1);
 }
+
+LINENUM
+atolnum(s)
+	char	*s;
+{
+	char	*os;
+	LINENUM	l = 0;
+	LINENUM	multmax;
+	LINENUM	lmax;
+	int	c;
+	int	neg = 0;
+
+	lmax = TYPE_MAXVAL(LINENUM);
+	multmax = TYPE_MAXVAL(LINENUM) / 10;
+
+	if (*s == '-') {
+		neg++;
+		s++;
+	} else if (*s == '+')
+		s++;
+
+	os = s;
+	while ((c = *s++) != '\0') {
+		if (c < '0' || c > '9')
+			break;
+		if (l > multmax) {
+			if (p_input_line)
+				malformed();
+			else
+				fatal(_("Number '%s' too large.\n"), os);
+		}
+		l *= 10;
+		c -= '0';
+		if (c > (lmax - l)) {
+			if (p_input_line)
+				malformed();
+			else
+				fatal(_("Number '%s' too large.\n"), os);
+		}
+		l += c;
+	}
+	if (s == ++os) {
+		if (p_input_line)
+			malformed();
+		else
+			fatal(_("Not a number '%s'.\n"), --os);
+	}
+	return (neg? -l:l);
+}
+
+int
+atoinum(s)
+	char	*s;
+{
+	LINENUM	l = atolnum(s);
+	int	ret;
+
+	ret = l;
+	if (ret != l)
+		fatal(_("Number '%lld' too large.\n"), (Llong)l);
+
+	return (ret);
+}
