@@ -29,10 +29,10 @@
 /*
  * Copyright 2006-2015 J. Schilling
  *
- * @(#)get.c	1.60 15/01/25 J. Schilling
+ * @(#)get.c	1.69 15/03/10 J. Schilling
  */
 #if defined(sun)
-#pragma ident "@(#)get.c 1.60 15/01/25 J. Schilling"
+#pragma ident "@(#)get.c 1.69 15/03/10 J. Schilling"
 #endif
 /*
  * @(#)get.c 1.59 06/12/12
@@ -55,13 +55,13 @@
 #define	DATELEN	12
 
 #if defined(__STDC__) || defined(PROTOTYPES)
-#define INCPATH		INS_BASE "/ccs/include"	/* place to find binaries */
+#define	INCPATH		INS_BASE "/ccs/include"	/* place to find binaries */
 #else
 /*
  * XXX With a K&R compiler, you need to edit the following string in case
  * XXX you like to change the install path.
  */
-#define INCPATH		"/usr/ccs/include"	/* place to find binaries */
+#define	INCPATH		"/usr/ccs/include"	/* place to find binaries */
 #endif
 
 
@@ -76,9 +76,9 @@
 
 static struct packet gpkt;		/* libcomobj data for get()	*/
 
+static Nparms	N;			/* Keep -N parameters		*/
 static struct sid sid;			/* To hold -r parameter		*/
 static unsigned	Ser;			/* To hold -a parameter		*/
-static char    *Whatstr = NULL;		/* To hold -w parameter		*/
 static char	*ilist;			/* To hold -i parameter		*/
 static char	*elist;			/* To hold -x parameter		*/
 static char	*lfile;			/* To hold -l parameter		*/
@@ -88,7 +88,7 @@ static char	gfile[PATH_MAX];	/* To hold -G parameter		*/
 static char	*Cwd = "";		/* To hold -C parameter		*/
 
 /* Beginning of modifications made for CMF system. */
-#define CMRLIMIT 128
+#define	CMRLIMIT 128
 static char	cmr[CMRLIMIT];		/* To hold -z parameter		*/
 /* End of insertion */
 
@@ -96,26 +96,16 @@ static int	num_files;		/* arg counter for main()/get()	*/
 static char	Pfilename[FILESIZE];	/* p.filename used in get()	*/
 static time_t	cutoff = MAX_TIME;	/* cutoff time for main()/get()	*/
 
-static int	num_ID_lines;		/* 's' flag for current file	*/
-static int	cnt_ID_lines;		/* current line cmp with above	*/
-static int	expand_IDs;		/* whether to expand ISs	*/
-static char	*list_expand_IDs;	/* 'y' flag for current file	*/
-static char	*Type;			/* 't' flag for current file	*/
-
 
 static	void    clean_up __PR((void));
 static	void	enter	__PR((struct packet *pkt, int ch, int n, struct sid *sidp));
 
 	int	main __PR((int argc, char **argv));
-static void	get __PR((char *file));
+static void	do_get __PR((char *file));
+static void	get __PR((struct packet *pkt, char *file));
 static void	gen_lfile __PR((struct packet *pkt));
-static void	idsetup __PR((struct packet *pkt));
-static void	makgdate __PR((char *old, char *new));
-static void	makgdatel __PR((char *old, char *new));
-static char *	idsubst __PR((struct packet *pkt, char *line));
-static char *	_trans __PR((char *tp, char *str, char *rest));
-static int	readcopy __PR((char *name, char *tp));
 static void	prfx __PR((struct packet *pkt));
+static void	annotate __PR((struct packet *pkt));
 static void	wrtpfile __PR((struct packet *pkt, char *inc, char *exc));
 static int	cmrinsert __PR((struct packet *pkt));
 
@@ -158,6 +148,9 @@ register char *argv[];
 #ifdef	SCCS_FATALHELP
 	Fflags |= FTLFUNC;
 	Ffunc = sccsfatalhelp;
+#endif
+#ifdef	XPG4
+	sccs_xpg4(TRUE);
 #endif
 	current_optind = 1;
 	optind = 1;
@@ -180,7 +173,7 @@ register char *argv[];
 			}
 			no_arg = 0;
 			i = current_optind;
-		        c = getopt(argc, argv, "-r:c:ebi:x:kl:Lpsmnogta:G:w:zqdC:FV(version)");
+		        c = getopt(argc, argv, "-r:c:ebi:x:kl:Lpsmnogta:G:w:zqdC:AFN:V(version)");
 				/* this takes care of options given after
 				** file names.
 				*/
@@ -201,8 +194,8 @@ register char *argv[];
 			}
 			p = optarg;
 			switch (c) {
-    			case CMFFLAG:
-				/* Concatenate the rest of this argument with 
+    			case CMFFLAG:		/* 'z' */
+				/* Concatenate the rest of this argument with
 				 * the existing CMR list. */
 				if (p) {
 				   while (*p) {
@@ -230,7 +223,7 @@ register char *argv[];
 				}
 				break;
 			case 'w':
-				if (p) Whatstr = p;
+				if (p) whatsetup(p);
 				break;
 			case 'c':
 				if (*p == 0) continue;
@@ -252,7 +245,7 @@ register char *argv[];
 				} else {
 				   no_arg = 1;
 				   lfile = NULL;
-				}	
+				}
 				break;
 			case 'i':
 				if (*p == 0) continue;
@@ -283,10 +276,11 @@ register char *argv[];
 			case 's':
 			case 't':
 			case 'd':
+			case 'A':
 			case 'F':
 				if (p) {
 				   sprintf(SccsError,
-				     gettext("value after %c arg (cm8)"), 
+				     gettext("value after %c arg (cm8)"),
 				     c);
 				   fatal(SccsError);
 				}
@@ -304,6 +298,15 @@ register char *argv[];
 				Cwd = p;
 				break;
 
+			case 'N':	/* Bulk names */
+				initN(&N);
+				if (optarg == argv[i+1]) {
+				   no_arg = 1;
+				   break;
+				}
+				N.n_parm = p;
+				break;
+
 			case 'V':		/* version */
 				printf("get %s-SCCS version %s %s (%s-%s-%s)\n",
 					PROVIDER,
@@ -313,8 +316,7 @@ register char *argv[];
 				exit(EX_OK);
 
 			default:
-			   fatal(gettext("Usage: get [-beFgkmLopst] [-l[p]] [-asequence] [-cdate-time] [-Gg-file] [-isid-list] [-rsid] [-xsid-list] file ..."));
-			
+			   fatal(gettext("Usage: get [-AbeFgkmLopst] [-l[p]] [-asequence] [-cdate-time] [-Gg-file]\n\t[-isid-list] [-rsid] [-xsid-list][ -N[bulk-spec]] file ..."));
 			}
 
 			/*
@@ -338,30 +340,46 @@ register char *argv[];
 		HADK = 1;
 	if (HADE && !logname())
 		fatal(gettext("User ID not in password file (cm9)"));
+	if (HADUCN) {					/* Parse -N args  */
+		parseN(&N);
+	}
 	setsig();
+	xsethome(NULL);
+	if (HADUCN && N.n_sdot && (sethomestat & SETHOME_OFFTREE))
+		fatal(gettext("-Ns. not supported in off-tree project mode"));
+
 	Fflags &= ~FTLEXIT;
 	Fflags |= FTLJMP;
 	for (i=1; i<argc; i++)
 		if ((p=argv[i]) != NULL)
-			do_file(p, get, 1, 1);
+			do_file(p, do_get, 1, N.n_sdot);
 
 	return (Fcnt ? 1 : 0);
 }
 
 static void
-get(file)
-char *file;
+do_get(file)
+	char		*file;
+{
+	get(&gpkt, file);
+}
+
+static void
+get(pkt, file)
+	struct packet	*pkt;
+	char		*file;
 {
 	register char *p;
 	register unsigned ser;
 	extern char had_dir, had_standinp;
 	struct stats stats;
-	char	str[32];		/* Must fit a SID string */
+	char	str[SID_STRSIZE];		/* Must fit a SID string */
 #ifdef	PROTOTYPES
 	char template[] = NOGETTEXT("/get.XXXXXX");
 #else
 	char *template = NOGETTEXT("/get.XXXXXX");
 #endif
+	char *ifile;
 	char buf1[PATH_MAX];
 	uid_t holduid;
 	gid_t holdgid;
@@ -369,6 +387,26 @@ char *file;
 
 	if (setjmp(Fjmp))
 		return;
+	if (HADUCN) {
+		char	*ofile = file;
+
+		file = bulkprepare(&N, file);
+		if (file == NULL) {
+			if (N.n_ifile)
+				ofile = N.n_ifile;
+			fatal(gettext("directory specified as s-file (cm14)"));
+		}
+		ifile = N.n_ifile;
+		if (sid.s_rel == 0 && N.n_sid.s_rel != 0) {
+			sid.s_rel = N.n_sid.s_rel;
+			sid.s_lev = N.n_sid.s_lev;
+			sid.s_br  = N.n_sid.s_br;
+			sid.s_seq = N.n_sid.s_seq;
+		}
+	} else {
+		ifile = NULL;
+	}
+
 	if (HADE) {
 		struct utsname	un;
 		char		*uuname;
@@ -379,123 +417,111 @@ char *file;
 		If it is, then create lock file.
 		pass both the PID and machine name to lockit
 		*/
-		sinit(&gpkt,file,0);
+		sinit(pkt, file, SI_INIT);
 		uname(&un);
 		uuname = un.nodename;
-		if (lockit(auxf(file,'z'),10, getpid(),uuname))
+		if (lockit(auxf(file, 'z'), 10, getpid(), uuname))
 			efatal(gettext("cannot create lock file (cm4)"));
 	}
 	/*
 	Open SCCS file and initialize packet
 	*/
-	sinit(&gpkt,file,1);
-	gpkt.p_ixuser = (HADI | HADX);
-	gpkt.p_reqsid.s_rel = sid.s_rel;
-	gpkt.p_reqsid.s_lev = sid.s_lev;
-	gpkt.p_reqsid.s_br = sid.s_br;
-	gpkt.p_reqsid.s_seq = sid.s_seq;
-	gpkt.p_verbose = (HADS) ? 0 : 1;
-	gpkt.p_stdout  = (HADP||lfile) ? stderr : stdout;
-	gpkt.p_cutoff = cutoff;
-	gpkt.p_lfile = lfile;
-	gpkt.p_enter = enter;
-	if ((gpkt.p_flags & PF_V6) == 0) {
-		gpkt.p_flags |= PF_GMT;
+	sinit(pkt, file, SI_OPEN);
+	pkt->p_ixuser = (HADI | HADX);
+	pkt->p_reqsid.s_rel = sid.s_rel;
+	pkt->p_reqsid.s_lev = sid.s_lev;
+	pkt->p_reqsid.s_br = sid.s_br;
+	pkt->p_reqsid.s_seq = sid.s_seq;
+	pkt->p_verbose = (HADS) ? 0 : 1;
+	pkt->p_stdout  = (HADP||lfile) ? stderr : stdout;
+	pkt->p_cutoff = cutoff;
+	pkt->p_lfile = lfile;
+	pkt->p_enter = enter;
+	if ((pkt->p_flags & PF_V6) == 0) {
+		pkt->p_flags |= PF_GMT;
 	} else if (cutoffstr != NULL) {
 		if (parse_date(cutoffstr, &cutoff, 0))
 			fatal(gettext("bad date/time (cm5)"));
-		gpkt.p_cutoff = cutoff;
+		pkt->p_cutoff = cutoff;
 	}
+	if (HADUCA)
+		pkt->p_pgmrs = (char **)Null;
 	if (Gfile[0] == 0 || !first) {
-		cat(gfile,Cwd,auxf(gpkt.p_file,'g'), (char *)0);
-		cat(Gfile,Cwd,auxf(gpkt.p_file,'A'), (char *)0);
+		gfile[0] = '\0';
+		strlcatl(gfile, sizeof (gfile),
+			Cwd,
+			ifile ? ifile :
+				auxf(pkt->p_file, 'g'),
+				(char *)0);
+		Gfile[0] = '\0';
+		strlcatl(Gfile, sizeof (Gfile),
+			Cwd,
+			ifile ? auxf(ifile, 'G') :
+				auxf(pkt->p_file, 'A'),
+			(char *)0);
 	}
 	strlcpy(buf1, dname(Gfile), sizeof (buf1));
 	strlcat(buf1, template, sizeof (buf1));
 	Gfile[0] = '\0';		/* File not yet created */
 	first = 0;
 
-	if (gpkt.p_verbose && (num_files > 1 || had_dir || had_standinp))
-		fprintf(gpkt.p_stdout,"\n%s:\n",gpkt.p_file);
-	if (dodelt(&gpkt,&stats,(struct sid *) 0,0) == 0)
-		fmterr(&gpkt);
-	finduser(&gpkt);
-	doflags(&gpkt);
-	if ((p = gpkt.p_sflags[SCANFLAG - 'a']) == NULL) {
-		num_ID_lines = 0;
-	} else {
-		num_ID_lines = atoi(p);
-	}
-	expand_IDs = 0;
-	if ((list_expand_IDs = gpkt.p_sflags[EXPANDFLAG - 'a']) != NULL) {
-		if (*list_expand_IDs) {
-			/*
-			 * The old Sun based code did break with get -k in case
-			 * that someone did previously call admin -fy...
-			 * Make sure that this now works correctly again.
-			 */
-			if (!HADK)		/* JS fix get -k */
-				expand_IDs = 1;
-		}
-	}
-	if (HADE) {
-		expand_IDs = 0;
-	}
-	if (!HADK) {
-		expand_IDs = 1;
-	}
+	if (pkt->p_verbose && (num_files > 1 || had_dir || had_standinp))
+		fprintf(pkt->p_stdout, "\n%s:\n", pkt->p_file);
+	if (dodelt(pkt, &stats, (struct sid *) 0, 0) == 0)
+		fmterr(pkt);
+	finduser(pkt);
+	doflags(pkt);
+
 	if (!HADA) {
-		ser = getser(&gpkt);
+		ser = getser(pkt);
 	} else {
-		if ((ser = Ser) > maxser(&gpkt))
+		if ((ser = Ser) > maxser(pkt))
 			fatal(gettext("serial number too large (ge19)"));
-		gpkt.p_gotsid = gpkt.p_idel[ser].i_sid;
-		if (HADR && sid.s_rel != gpkt.p_gotsid.s_rel) {
-			zero((char *) &gpkt.p_reqsid, sizeof(gpkt.p_reqsid));
-			gpkt.p_reqsid.s_rel = sid.s_rel;
+		pkt->p_gotsid = pkt->p_idel[ser].i_sid;
+		if (HADR && sid.s_rel != pkt->p_gotsid.s_rel) {
+			zero((char *) &pkt->p_reqsid, sizeof (pkt->p_reqsid));
+			pkt->p_reqsid.s_rel = sid.s_rel;
 		} else {
-			gpkt.p_reqsid = gpkt.p_gotsid;
+			pkt->p_reqsid = pkt->p_gotsid;
 		}
 	}
-	doie(&gpkt,ilist,elist,(char *) 0);
-	setup(&gpkt,(int) ser);
-	if ((Type = gpkt.p_sflags[TYPEFLAG - 'a']) == NULL)
-		Type = Null;
+	doie(pkt, ilist, elist, (char *) 0);
+	setup(pkt, (int) ser);
 	if (!(HADP || HADG)) {
 		if (exists(gfile) && (S_IWRITE & Statbuf.st_mode)) {
-			sprintf(SccsError,gettext("writable `%s' exists (ge4)"),
+			sprintf(SccsError, gettext("writable `%s' exists (ge4)"),
 				gfile);
 			fatal(SccsError);
 		}
 	}
-	if (gpkt.p_verbose) {
-		sid_ba(&gpkt.p_gotsid,str);
-		fprintf(gpkt.p_stdout,"%s\n",str);
+	if (pkt->p_verbose) {
+		sid_ba(&pkt->p_gotsid, str);
+		fprintf(pkt->p_stdout, "%s\n", str);
 	}
 	if (HADE) {
-		if (HADC || gpkt.p_reqsid.s_rel == 0)
-			gpkt.p_reqsid = gpkt.p_gotsid;
-		newsid(&gpkt, gpkt.p_sflags[BRCHFLAG - 'a'] && HADB);
-		permiss(&gpkt);
-		wrtpfile(&gpkt,ilist,elist);
+		if (HADC || pkt->p_reqsid.s_rel == 0)
+			pkt->p_reqsid = pkt->p_gotsid;
+		newsid(pkt, pkt->p_sflags[BRCHFLAG - 'a'] && HADB);
+		permiss(pkt);
+		wrtpfile(pkt, ilist, elist);
 	}
 	if (!HADG || HADL) {
-		if (gpkt.p_stdout) {
-			fflush(gpkt.p_stdout);
-			fflush (stderr);
+		if (pkt->p_stdout) {
+			fflush(pkt->p_stdout);
+			fflush(stderr);
 		}
 		holduid=geteuid();
 		holdgid=getegid();
 		setuid(getuid());
 		setgid(getgid());
 		if (HADL)
-			gen_lfile(&gpkt);
+			gen_lfile(pkt);
 		if (HADG)
 			goto unlock;
-		flushto(&gpkt,EUSERTXT,1);
-		idsetup(&gpkt);
-		gpkt.p_chkeof = 1;
-		gpkt.p_did_id = 0;
+		flushto(pkt, EUSERTXT, 1);
+		idsetup(pkt, gfile);
+		pkt->p_chkeof = 1;
+		pkt->p_did_id = 0;
 		/*
 		call `xcreate' which deletes the old `g-file' and
 		creates a new one except if the `p' keyletter is set in
@@ -503,9 +529,9 @@ char *file;
 		The mod of the file depends on whether or not the `k'
 		keyletter has been set.
 		*/
-		if (gpkt.p_gout == 0) {
+		if (pkt->p_gout == 0) {
 			if (HADP) {
-				gpkt.p_gout = stdout;
+				pkt->p_gout = stdout;
 			} else {
 #ifdef	HAVE_MKSTEMP
 				close(mkstemp(buf1));	/* safer than mktemp */
@@ -518,12 +544,12 @@ char *file;
 				 * instead of xfcreat() in order to avoid an
 				 * unlink()/create() chain.
 				 */
-				if ((exists(gpkt.p_file) && (S_IEXEC & Statbuf.st_mode)) ||
-				    (gpkt.p_flags & PF_SCOX)) {
-					gpkt.p_gout = xfcreat(Gfile,HADK ? 
+				if ((exists(pkt->p_file) && (S_IEXEC & Statbuf.st_mode)) ||
+				    (pkt->p_flags & PF_SCOX)) {
+					pkt->p_gout = xfcreat(Gfile, HADK ?
 						((mode_t)0755) : ((mode_t)0555));
 				} else {
-					gpkt.p_gout = xfcreat(Gfile,HADK ? 
+					pkt->p_gout = xfcreat(Gfile, HADK ?
 						((mode_t)0644) : ((mode_t)0444));
 				}
 #ifdef	USE_SETVBUF
@@ -531,78 +557,78 @@ char *file;
 				 * Do not call setvbuf() with stdout as this may result
 				 * in a second illegal call in gen_lfile().
 				 */
-				setvbuf(gpkt.p_gout, NULL, _IOFBF, VBUF_SIZE);
+				setvbuf(pkt->p_gout, NULL, _IOFBF, VBUF_SIZE);
 #endif
 			}
 		}
-		gpkt.p_ghash = 0;		/* Reset ghash */
-		if (gpkt.p_encoding & EF_UUENCODE) {
-			while (readmod(&gpkt)) {
-				decode(gpkt.p_line,gpkt.p_gout);
+		pkt->p_ghash = 0;		/* Reset ghash */
+		if (pkt->p_encoding & EF_UUENCODE) {
+			while (readmod(pkt)) {
+				decode(pkt->p_line, pkt->p_gout);
 			}
 		} else {
-			while (readmod(&gpkt)) {
-				prfx(&gpkt);
-				if (gpkt.p_flags & PF_NONL)
-					gpkt.p_line[gpkt.p_line_length-1] = '\0';
-				p = idsubst(&gpkt, gpkt.p_lineptr);
-				if (fputs(p,gpkt.p_gout)==EOF)
+			while (readmod(pkt)) {
+				prfx(pkt);
+				if (pkt->p_flags & PF_NONL)
+					pkt->p_line[pkt->p_line_length-1] = '\0';
+				p = idsubst(pkt, pkt->p_lineptr);
+				if (fputs(p, pkt->p_gout) == EOF)
 					xmsg(gfile, NOGETTEXT("get"));
 			}
 		}
-		if ((gpkt.p_flags & (PF_V6 | PF_V6TAGS)) && gpkt.p_hash) {
+		if ((pkt->p_flags & (PF_V6 | PF_V6TAGS)) && pkt->p_hash) {
 			/*
 			 * SCCS v6 is able to check against SID specific
 			 * checksums, but this will not work in case that
 			 * we use include or exclude lists.
 			 */
-			if (gpkt.p_hash[ser] != (gpkt.p_ghash & 0xFFFF) &&
+			if (pkt->p_hash[ser] != (pkt->p_ghash & 0xFFFF) &&
 			    elist == NULL && ilist == NULL && !HADUCF)
 				fatal(gettext("corrupted file version (co27)"));
 		}
-			
-		if (gpkt.p_gout) {
-			if (fflush(gpkt.p_gout) == EOF)
+
+		if (pkt->p_gout) {
+			if (fflush(pkt->p_gout) == EOF)
 				xmsg(gfile, NOGETTEXT("get"));
 			fflush (stderr);
 		}
-		if (gpkt.p_gout && gpkt.p_gout != stdout) {
+		if (pkt->p_gout && pkt->p_gout != stdout) {
 			/*
 			 * Force g-file to disk and verify
 			 * that it actually got there.
 			 */
 #ifdef	HAVE_FSYNC
-			if (fsync(fileno(gpkt.p_gout)) < 0)
+			if (fsync(fileno(pkt->p_gout)) < 0)
 				xmsg(gfile, NOGETTEXT("get"));
 #endif
-			if (fclose(gpkt.p_gout) == EOF)
+			if (fclose(pkt->p_gout) == EOF)
 				xmsg(gfile, NOGETTEXT("get"));
-			gpkt.p_gout = NULL;
+			pkt->p_gout = NULL;
 		}
-		if (gpkt.p_verbose) {
-#ifdef XPG4		
-		   fprintf(gpkt.p_stdout, NOGETTEXT("%d lines\n"), gpkt.p_glnno);
+		if (pkt->p_verbose) {
+#ifdef XPG4
+		   fprintf(pkt->p_stdout, NOGETTEXT("%d lines\n"), pkt->p_glnno);
 #else
 		   if (HADD == 0)
-		      fprintf(gpkt.p_stdout,gettext("%d lines\n"),gpkt.p_glnno);
+		      fprintf(pkt->p_stdout, gettext("%d lines\n"), pkt->p_glnno);
 #endif
-		}		
-		if (!gpkt.p_did_id && !HADK && !HADQ &&
-		    (!gpkt.p_sflags[EXPANDFLAG - 'a'] ||
-		    *(gpkt.p_sflags[EXPANDFLAG - 'a']))) {
-		   if (gpkt.p_sflags[IDFLAG - 'a']) {
-		      if (!(*gpkt.p_sflags[IDFLAG - 'a'])) {
+		}
+		if (!pkt->p_did_id && !HADK && !HADQ &&
+		    (!pkt->p_sflags[EXPANDFLAG - 'a'] ||
+		    *(pkt->p_sflags[EXPANDFLAG - 'a']))) {
+		   if (pkt->p_sflags[IDFLAG - 'a']) {
+		      if (!(*pkt->p_sflags[IDFLAG - 'a'])) {
 		         fatal(gettext("no id keywords (cm6)"));
 		      } else {
 			 fatal(gettext("invalid id keywords (cm10)"));
 		      }
 		   } else {
-		      if (gpkt.p_verbose) {
+		      if (pkt->p_verbose) {
 			 fprintf(stderr, gettext("No id keywords (cm7)\n"));
 		      }
 		   }
 		}
-		if (Gfile[0] != '\0' && exists(Gfile) ) {
+		if (Gfile[0] != '\0' && exists(Gfile)) {
 			rename(Gfile, gfile);
 			if (HADO) {
 				struct tm	tm;
@@ -610,18 +636,18 @@ char *file;
 				unsigned int	gser;
 				extern dtime_t	Timenow;
 
-				gser = sidtoser(&gpkt.p_gotsid, &gpkt);
+				gser = sidtoser(&pkt->p_gotsid, pkt);
 
 				ts[0].tv_sec = Timenow.dt_sec;
 				ts[0].tv_nsec = Timenow.dt_nsec;
-				ts[1].tv_sec = gpkt.p_idel[gser].i_datetime.tv_sec;
-				ts[1].tv_nsec = gpkt.p_idel[gser].i_datetime.tv_nsec;
+				ts[1].tv_sec = pkt->p_idel[gser].i_datetime.tv_sec;
+				ts[1].tv_nsec = pkt->p_idel[gser].i_datetime.tv_nsec;
 				/*
 				 * If we did cheat while scanning the delta
 				 * table and converted the time stamps assuming
 				 * GMT. Fix the resulting error here.
 				 */
-				if (gpkt.p_flags & PF_GMT) {
+				if (pkt->p_flags & PF_GMT) {
 					tm = *gmtime(&ts[1].tv_sec);
 					tm.tm_isdst = -1;
 					ts[1].tv_sec = mktime(&tm);
@@ -632,20 +658,20 @@ char *file;
 		setuid(holduid);
 		setgid(holdgid);
 	}
-	if (gpkt.p_iop) {
-		fclose(gpkt.p_iop);
-		gpkt.p_iop = NULL;
+	if (pkt->p_iop) {
+		fclose(pkt->p_iop);
+		pkt->p_iop = NULL;
 	}
 unlock:
 	if (HADE) {
 		struct utsname	un;
 		char		*uuname;
 
-		copy(auxf(gpkt.p_file,'p'),Pfilename);
-		rename(auxf(gpkt.p_file,'q'),Pfilename);
+		copy(auxf(pkt->p_file, 'p'), Pfilename);
+		rename(auxf(pkt->p_file, 'q'), Pfilename);
 		uname(&un);
 		uuname = un.nodename;
-		unlockit(auxf(gpkt.p_file,'z'), getpid(),uuname);
+		unlockit(auxf(pkt->p_file, 'z'), getpid(), uuname);
 	}
 	ffreeall();
 }
@@ -657,7 +683,7 @@ char ch;
 int n;
 struct sid *sidp;
 {
-	char	str[32];		/* Must fit a SID string */
+	char	str[SID_STRSIZE];		/* Must fit a SID string */
 	register struct apply *ap;
 
 	sid_ba(sidp,str);
@@ -694,7 +720,7 @@ register struct packet *pkt;
 {
 	char *n;
 	int reason;
-	char str[DT_ZSTRSIZE];		/* SCCS v6 date or SID str */
+	char str[max(DT_ZSTRSIZE, SID_STRSIZE)]; /* SCCS v6 date or SID str */
 	char line[BUFSIZ];
 	struct deltab dt;
 	FILE *in;
@@ -735,7 +761,7 @@ register struct packet *pkt;
 				}
 			}
 			switch (reason & (INCL | EXCL | CUTOFF)) {
-	
+
 			case INCL:
 				OUTPUTC('I');
 				break;
@@ -784,7 +810,7 @@ register struct packet *pkt;
 				case COMMENTS:
 					if (dt.d_type == 'D') {
 					   if (fprintf(out,"\t%s",&line[3]) == EOF)
-					      xmsg(outname, 
+					      xmsg(outname,
 					         NOGETTEXT("gen_lfile"));
 					}
 					continue;
@@ -809,476 +835,39 @@ register struct packet *pkt;
 #undef	OUTPUTC
 }
 
-static char	Curdatel[DT_ZSTRSIZE];
-static char	Curdate[DT_ZSTRSIZE];
-static char	*Curtime;
-static char	Gdate[DATELEN];
-static char	Gdatel[DATELEN];
-static char	Chgdate[DT_ZSTRSIZE];
-static char	Chgdatel[DT_ZSTRSIZE];
-static char	*Chgtime;
-static char	Gchgdate[DATELEN];
-static char	Gchgdatel[DATELEN];
-static char	Sid[32];
-static char	Mod[FILESIZE];
-static char	Olddir[BUFSIZ];
-static char	Pname[BUFSIZ];
-static char	Dir[BUFSIZ];
-static char	*Qsect;
-
-static void
-idsetup(pkt)
-register struct packet *pkt;
-{
-	extern dtime_t Timenow;
-	register int n;
-	register char *p;
-
-	date_ba(&Timenow.dt_sec, Curdate, 0);
-	date_bal(&Timenow.dt_sec, Curdatel, 0);
-	Curdatel[10] = 0;
-	Curtime = &Curdate[9];
-	Curdate[8] = 0;
-	makgdate(Curdate,Gdate);
-	makgdatel(Curdatel,Gdatel);
-	for (n = maxser(pkt); n; n--)
-		if (pkt->p_apply[n].a_code == APPLY)
-			break;
-	if (n) {
-		date_ba(&pkt->p_idel[n].i_datetime.tv_sec, Chgdate, pkt->p_flags);
-		date_bal(&pkt->p_idel[n].i_datetime.tv_sec, Chgdatel, pkt->p_flags);
-	} else {
-#ifdef XPG4
-		FILE	*xf;
-
-		if (exists(gfile))
-			unlink(gfile);
-		xf = xfcreat(gfile, HADK ? ((mode_t)0644) : ((mode_t)0444));
-		if (xf)
-			fclose(xf);
-#endif
-		fatal(gettext("No sid prior to cutoff date-time (ge23)"));
-	}
-	Chgdatel[10] = 0;
-	Chgtime = &Chgdate[9];
-	Chgdate[8] = 0;
-	makgdate(Chgdate,Gchgdate);
-	makgdatel(Chgdatel,Gchgdatel);
-	sid_ba(&pkt->p_gotsid,Sid);
-	if ((p = pkt->p_sflags[MODFLAG - 'a']) != NULL)
-		copy(p,Mod);
-	else
-		copy(auxf(gpkt.p_file,'g'), Mod);
-	if ((Qsect = pkt->p_sflags[QSECTFLAG - 'a']) == NULL)
-		Qsect = Null;
-}
-
-#ifdef	HAVE_STRFTIME
-static void 
-makgdate(old,new)
-register char *old, *new;
-{
-	struct tm	tm;
-	sscanf(old, "%d/%d/%d", &(tm.tm_year), &(tm.tm_mon), &(tm.tm_mday));
-	tm.tm_mon--;
-	strftime (new, (size_t) DATELEN, NOGETTEXT("%D"), &tm);
-}
-
-static void 
-makgdatel(old,new)
-register char *old, *new;
-{
-	struct tm	tm;
-	sscanf(old, "%d/%d/%d", &(tm.tm_year), &(tm.tm_mon), &(tm.tm_mday));
-	if (tm.tm_year > 1900)
-		tm.tm_year -= 1900;
-	tm.tm_mon--;
-	strftime (new, (size_t) DATELEN, NOGETTEXT("%m/%d/%Y"), &tm);
-}
-	
-
-#else	/* HAVE_STRFTIME not defined */
-
-static void
-makgdate(old,new)
-register char *old, *new;
-{
-#define	NULL_PAD_DATE		/* Make it compatible to strftime()/man page */
-#ifndef	NULL_PAD_DATE
-	if ((*new = old[3]) != '0')
-		new++;
-#else
-	*new++ = old[3];	
-#endif
-	*new++ = old[4];
-	*new++ = '/';
-#ifndef	NULL_PAD_DATE
-	if ((*new = old[6]) != '0')
-		new++;
-#else
-	*new++ = old[6];
-#endif
-	*new++ = old[7];
-	*new++ = '/';
-	*new++ = old[0];
-	*new++ = old[1];
-	*new = 0;
-}
-
-static void
-makgdatel(old,new)
-register char *old, *new;
-{
-	char	*p;
-	int	off;
-
-	for (p = old; *p >= '0' && *p <= '9'; p++)
-		;
-	off = p - old - 2;
-	if (off < 0)
-		off = 0;
-	
-#ifndef	NULL_PAD_DATE
-	if ((*new = old[3+off]) != '0')
-		new++;
-#else
-	*new++ = old[3+off];
-#endif
-	*new++ = old[4+off];
-	*new++ = '/';
-#ifndef	NULL_PAD_DATE
-	if ((*new = old[6+off]) != '0')
-		new++;
-#else
-	*new++ = old[6+off];
-#endif
-	*new++ = old[7+off];
-	*new++ = '/';
-	*new++ = old[0];
-	*new++ = old[1];
-	if (off > 0)
-		*new++ = old[2];
-	if (off > 1)
-		*new++ = old[3];
-	*new = 0;
-}
-#endif	/* HAVE_STRFTIME */
-
-
-static char Zkeywd[5] = "@(#)";
-static char *tline = NULL;
-static size_t tline_size = 0;
-
-#define trans(a, b)	_trans(a, b, lp)
-
-static char *
-idsubst(pkt,line)
-register struct packet *pkt;
-char line[];
-{
-	static char *hold = NULL;
-	static size_t hold_size = 0;
-	size_t new_hold_size;
-	static char str[32];
-	register char *lp, *tp;
-	int recursive = 0;
-	int expand_XIDs;
-	char *expand_ID;
-	register char **sflags = pkt->p_sflags;
-		
-	cnt_ID_lines++;
-	if (HADK) {
-		if (!expand_IDs)
-			return(line);
-	} else {
-		if (!any('%', line))
-			return(line);
-	}
-	if (cnt_ID_lines > num_ID_lines) {
-		if (!expand_IDs) {
-			return(line);
-		}
-		if (num_ID_lines) {
-			expand_IDs = 0;
-			return(line);
-		}
-	}
-	expand_XIDs = sflags[EXTENSFLAG - 'a'] != NULL || list_expand_IDs != NULL;
-	tp = tline;
-	for (lp=line; *lp != 0; lp++) {
-		if ((!pkt->p_did_id) && (sflags[IDFLAG - 'a']) &&
-		    (!(strncmp(sflags[IDFLAG-'a'],lp,strlen(sflags[IDFLAG-'a'])))))
-				pkt->p_did_id = 1;
-		if (lp[0] == '%' && lp[1] != 0 && lp[2] == '%') {
-			expand_ID = ++lp;
-			if (!HADK && expand_IDs && (list_expand_IDs != NULL)) {
-				if (!any(*expand_ID, list_expand_IDs)) {
-					str[3] = '\0';
-					strncpy(str, lp - 1, 3);
-					tp = trans(tp, str);
-					lp++;
-					continue;
-				}
-			}
-			switch (*expand_ID) {
-
-			case 'M':
-				tp = trans(tp,Mod);
-				break;
-			case 'Q':
-				tp = trans(tp,Qsect);
-				break;
-			case 'R':
-				sprintf(str,"%d",pkt->p_gotsid.s_rel);
-				tp = trans(tp,str);
-				break;
-			case 'L':
-				sprintf(str,"%d",pkt->p_gotsid.s_lev);
-				tp = trans(tp,str);
-				break;
-			case 'B':
-				sprintf(str,"%d",pkt->p_gotsid.s_br);
-				tp = trans(tp,str);
-				break;
-			case 'S':
-				sprintf(str,"%d",pkt->p_gotsid.s_seq);
-				tp = trans(tp,str);
-				break;
-			case 'D':
-				tp = trans(tp,Curdate);
-				break;
-			case 'd':
-				if (!expand_XIDs)
-					goto noexpand;
-				tp = trans(tp,Curdatel);
-				break;
-			case 'H':
-				tp = trans(tp,Gdate);
-				break;
-			case 'h':
-				if (!expand_XIDs)
-					goto noexpand;
-				tp = trans(tp,Gdatel);
-				break;
-			case 'T':
-				tp = trans(tp,Curtime);
-				break;
-			case 'E':
-				tp = trans(tp,Chgdate);
-				break;
-			case 'e':
-				if (!expand_XIDs)
-					goto noexpand;
-				tp = trans(tp,Chgdatel);
-				break;
-			case 'G':
-				tp = trans(tp,Gchgdate);
-				break;
-			case 'g':
-				if (!expand_XIDs)
-					goto noexpand;
-				tp = trans(tp,Gchgdatel);
-				break;
-			case 'U':
-				tp = trans(tp,Chgtime);
-				break;
-			case 'Z':
-				tp = trans(tp,Zkeywd);
-				break;
-			case 'Y':
-				tp = trans(tp,Type);
-				break;
-			/*FALLTHRU*/
-			case 'W':
-				if ((Whatstr != NULL) && (recursive == 0)) {
-					recursive = 1;
-					lp += 2;
-					new_hold_size = strlen(lp) + strlen(Whatstr) + 1;
-					if (new_hold_size > hold_size) {
-						if (hold)
-							free(hold);
-						hold_size = new_hold_size;
-						hold = (char *)malloc(hold_size);
-						if (hold == NULL) {
-							efatal(gettext("OUT OF SPACE (ut9)"));
-						}
-					}
-					strcpy(hold,Whatstr);
-					strcat(hold,lp);
-					lp = hold;
-					lp--;
-					continue;
-				}
-				tp = trans(tp,Zkeywd);
-				tp = trans(tp,Mod);
-				tp = trans(tp,"\t");
-			case 'I':
-				tp = trans(tp,Sid);
-				break;
-			case 'P':
-				copy(pkt->p_file,Dir);
-				dname(Dir);
-				if (getcwd(Olddir,sizeof(Olddir)) == NULL)
-				   efatal(gettext("getcwd() failed (ge20)"));
-				if (chdir(Dir) != 0)
-				   efatal(gettext("cannot change directory (ge21)"));
-			  	if (getcwd(Pname,sizeof(Pname)) == NULL)
-				   efatal(gettext("getcwd() failed (ge20)"));
-				if (chdir(Olddir) != 0)
-				   efatal(gettext("cannot change directory (ge21)"));
-				tp = trans(tp,Pname);
-				tp = trans(tp,"/");
-				tp = trans(tp,(sname(pkt->p_file)));
-				break;
-			case 'F':
-				tp = trans(tp,pkt->p_file);
-				break;
-			case 'C':
-				sprintf(str,"%d",pkt->p_glnno);
-				tp = trans(tp,str);
-				break;
-			case 'A':
-				tp = trans(tp,Zkeywd);
-				tp = trans(tp,Type);
-				tp = trans(tp," ");
-				tp = trans(tp,Mod);
-				tp = trans(tp," ");
-				tp = trans(tp,Sid);
-				tp = trans(tp,Zkeywd);
-				break;
-			noexpand:
-			default:
-				str[0] = '%';
-				str[1] = *lp;
-				str[2] = '\0';
-				tp = trans(tp,str);
-				continue;
-			}
-			if (!(sflags[IDFLAG - 'a']))
-				pkt->p_did_id = 1;
-			lp++;
-		} else if (strncmp(lp, "%sccs.include.", 14) == 0) {
-			register char *p;
-
-			for (p = lp + 14; *p && *p != '%'; ++p);
-			if (*p == '%') {
-				*p = '\0';
-				if (readcopy(lp + 14, tline))
-					return(tline);
-				*p = '%';
-			}
-			str[0] = *lp;
-			str[1] = '\0';
-			tp = trans(tp,str);
-		} else {
-			str[0] = *lp;
-			str[1] = '\0';
-			tp = trans(tp,str);
-		}
-	}
-	return(tline);
-}
-
-static char *
-_trans(tp,str,rest)
-register char *tp, *str, *rest;
-{
-	register size_t filled_size	= tp - tline;
-	register size_t new_size	= filled_size + strlen(str) + 1;
-
-	if (new_size > tline_size) {
-		tline_size = new_size + strlen(rest) + DEF_LINE_SIZE;
-		tline = (char*) realloc(tline, tline_size);
-		if (tline == NULL) {
-			efatal(gettext("OUT OF SPACE (ut9)"));
-		}
-		tp = tline + filled_size;
-	}
-	while((*tp++ = *str++) != '\0')
-		;
-	return(tp-1);
-}
-
-#ifndef	HAVE_STRERROR
-#define	strerror(p)		errmsgstr(p)	/* Use libschily version */
-#endif
-static int
-readcopy(name, tp)
-	register char *name;
-	register char *tp;
-{
-	register FILE *fp = NULL;
-	register int ch;
-	char path[MAXPATHLEN];
-	struct stat sb;
-	register size_t filled_size	= tp - tline;
-	register size_t new_size	= filled_size;
-
-	if (!HADK && expand_IDs && (list_expand_IDs != NULL)) {
-		if (!any('*', list_expand_IDs))
-			return (0);
-	}
-	if (fp == NULL) {
-		char *np;
-		char *p = getenv(NOGETTEXT("SCCS_INCLUDEPATH"));
-
-		if (p == NULL)
-			p = INCPATH;
-		do {
-			for (np = p; *np; np++)
-				if (*np == ':')
-					break;
-			(void)snprintf(path, sizeof (path), "%.*s/%s",
-						(int)(np-p), p, name);
-			if (p == np || *p == '\0')
-				strlcpy(path, name, sizeof (path));
-			fp = fopen(path, "r");
-			p = np;
-			if (*p == '\0')
-				break;
-			if (*p == ':')
-				p++;
-		} while (fp == NULL);
-	}
-	if (fp == NULL) {
-		fprintf(stderr, gettext("WARNING: %s: %s.\n"),
-						path, strerror(errno));
-		return (0);
-	}
-	if (fstat(fileno(fp), &sb) < 0)
-		return (0);
-	new_size += sb.st_size + 1;
-
-	if (new_size > tline_size) {
-		tline_size = new_size + DEF_LINE_SIZE;
-		tline = (char*) realloc(tline, tline_size);
-		if (tline == NULL) {
-			efatal(gettext("OUT OF SPACE (ut9)"));
-		}
-		tp = tline + filled_size;
-	}
-
-	while ((ch = getc(fp)) != EOF)
-		*tp++ = ch;
-	*tp++ = '\0';
-	(void)fclose(fp);
-	return (1);
-}
-
 static void
 prfx(pkt)
 register struct packet *pkt;
 {
-	char str[32];		/* Must fit a SID string */
+	char str[SID_STRSIZE];		/* Must fit a SID string */
 
 	if (HADN)
-		if (fprintf(pkt->p_gout, "%s\t", Mod) == EOF)
+		if (fprintf(pkt->p_gout, "%s\t", getmodname()) == EOF)
 			xmsg(gfile, NOGETTEXT("prfx"));
 	if (HADM) {
 		sid_ba(&pkt->p_inssid,str);
 		if (fprintf(pkt->p_gout, "%s\t", str) == EOF)
 			xmsg(gfile, NOGETTEXT("prfx"));
 	}
+	if (pkt->p_pgmrs)
+		annotate(pkt);
+}
+
+static void
+annotate(pkt)
+	register struct packet *pkt;
+{
+	char	str[SID_STRSIZE];	/* Must fit a SID string */
+	char	*p;
+
+	date_ba(&pkt->p_idel[pkt->p_insser].i_datetime.tv_sec,
+				str, pkt->p_flags);
+	p = strchr(str, ' ');
+	if (p)
+		*p = '\0';
+	if (fprintf(pkt->p_gout, "%-8s\t%s\t",
+				pkt->p_pgmrs[pkt->p_insser], str) == EOF)
+		xmsg(gfile, NOGETTEXT("prfx"));
 }
 
 static void
@@ -1322,7 +911,7 @@ register struct packet *pkt;
 char *inc, *exc;
 {
 	char line[64];
-	char str1[32], str2[32];	/* Must fit a SID string */
+	char str1[SID_STRSIZE], str2[SID_STRSIZE]; /* Must fit a SID string */
 	char *user, *pfile;
 	FILE *in, *out;
 	struct pfile pf;
@@ -1375,7 +964,7 @@ char *inc, *exc;
 	sid_ba(&pkt->p_gotsid,str1);
 	sid_ba(&pkt->p_reqsid,str2);
 	/*
-	 * POSIX does not explicitly mention a 2-digit year for the p-file but 
+	 * POSIX does not explicitly mention a 2-digit year for the p-file but
 	 * refers to "date-time" which is most likely expected to have 2-digits.
 	 * We use 4-digits before 1969 and past 2068
 	 * which is outside the year range specified by POSIX.
