@@ -2,11 +2,13 @@
  * CDDL HEADER START
  *
  * The contents of this file are subject to the terms of the
- * Common Development and Distribution License (the "License").
- * You may not use this file except in compliance with the License.
+ * Common Development and Distribution License ("CDDL"), version 1.0.
+ * You may only use this file in accordance with the terms of version
+ * 1.0 of the CDDL.
  *
- * You can obtain a copy of the license at usr/src/OPENSOLARIS.LICENSE
- * or http://www.opensolaris.org/os/licensing.
+ * A full copy of the text of the CDDL should have accompanied this
+ * source.  A copy of the CDDL is also available via the Internet at
+ * http://www.opensource.org/licenses/cddl1.txt
  * See the License for the specific language governing permissions
  * and limitations under the License.
  *
@@ -25,14 +27,14 @@
  * Use is subject to license terms.
  */
 /*
- * This file contains modifications Copyright 2008-2011 J. Schilling
+ * This file contains modifications Copyright 2008-2015 J. Schilling
  *
- * @(#)date_bal.c	1.8 11/08/24 J. Schilling
+ * @(#)date_bal.c	1.11 15/01/26 J. Schilling
  *
  * From Sun: @(#)sccs:lib/comobj/date_ba.c @(#)date_ba.c 1.5 06/12/12
  */
 #if defined(sun)
-#pragma ident "@(#)date_bal.c 1.8 11/08/24 J. Schilling"
+#pragma ident "@(#)date_bal.c 1.11 15/01/26 J. Schilling"
 #endif
 # include	<defines.h>
 
@@ -65,6 +67,7 @@ int	flags;
 	register struct tm *lcltm;
 	register char	*p;
 		int	zone = bdt->dt_zone;
+		int	nsec = bdt->dt_nsec;
 
 #if defined(BUG_1205145) || defined(GMT_TIME)
 	lcltm = gmtime(&bdt->dt_sec);
@@ -90,15 +93,34 @@ int	flags;
 	DO2(p,lcltm->tm_hour,':');
 	DO2(p,lcltm->tm_min,':');
 	DO2(p,lcltm->tm_sec,0);
+	if (nsec > 0 && nsec < 1000000000) {
+		char	*psave;
+		int	n = 10;
+
+		if (nsec % 1000 == 0) {
+			nsec /= 1000;
+			n = 7;
+		}
+		*--p = '.';
+		psave = p += n;
+		*p = '\0';
+		while (--n > 0) {
+			*--p = '0' + (nsec % 10);
+			nsec /= 10;
+		}
+		p = ++psave;
+	}
 	if (zone != DT_NO_ZONE) {
 		register int	z = zone / 60;	/* seconds -> minutes */
 		register int	n;
 
 		--p;
-		if (z < 0)
+		if (z < 0) {
 			*p++ = '-';
-		else
+			z = -z;
+		} else {
 			*p++ = '+';
+		}
 		n = z / 60;
 		DO2_(p, n);
 		n = z % 60;
