@@ -10,10 +10,10 @@
  * file and include the License file CDDL.Schily.txt from this distribution.
  */
 /*
- * @(#)putmeta.c	1.1 11/11/01 Copyright 2011 J. Schilling
+ * @(#)putmeta.c	1.3 15/02/25 Copyright 2011-2015 J. Schilling
  */
 #if defined(sun)
-#pragma ident "@(#)putmeta.c	1.1 11/11/01 Copyright 2011 J. Schilling"
+#pragma ident "@(#)putmeta.c	1.3 15/02/25 Copyright 2011-2015 J. Schilling"
 #endif
 
 #if defined(sun)
@@ -27,6 +27,7 @@ putmeta(pkt)
 	register struct packet	*pkt;
 {
 	char	line[max(8192, PATH_MAX+1)];
+	char	urbuf[20];
 
 	if (pkt->p_init_path) {
 		snprintf(line, sizeof (line), "%c%c %s %s\n",
@@ -34,19 +35,8 @@ putmeta(pkt)
 			pkt->p_init_path);
 		putline(pkt, line);
 	}
-#ifdef	HAVE_LONG_LONG
-	if (pkt->p_rand != 0) {
-		snprintf(line, sizeof (line), "%c%c %s %llx\n",
-			CTLCHAR, GLOBALEXTENS, "r",
-			pkt->p_rand);
-		putline(pkt, line);
-	}
-#else
-	if (pkt->p_rand.high != 0 || pkt->p_rand.low != 0) {
-		snprintf(line, sizeof (line), "%c%c %s %x%8.8x\n",
-			CTLCHAR, GLOBALEXTENS, "r",
-			pkt->p_rand.high, pkt->p_rand.low);
-		putline(pkt, line);
-	}
-#endif
+	urand_ba(&pkt->p_rand, urbuf, sizeof (urbuf));
+	snprintf(line, sizeof (line), "%c%c %s %s\n",
+		CTLCHAR, GLOBALEXTENS, "r", urbuf);
+	putline(pkt, line);
 }
