@@ -1,4 +1,8 @@
-h07128
+h17481
+s 00001/00001/00037
+d D 1.2 15/06/03 00:06:43 joerg 2 1
+c ../common/test-common -> ../../common/test-common
+e
 s 00038/00000/00000
 d D 1.1 10/05/11 11:30:00 joerg 1 0
 c date and time created 10/05/11 11:30:00 by joerg
@@ -15,7 +19,12 @@
 # t-option.sh:  Testing for correct operation of admin -t.
 
 # Import common functions & definitions.
+D 2
 . ../common/test-common
+E 2
+I 2
+. ../../common/test-common
+E 2
 
 expands_to () {
     # $1 -- label
