@@ -1,4 +1,8 @@
-h16085
+h26436
+s 00001/00001/00052
+d D 1.2 15/06/03 00:06:44 joerg 2 1
+c ../common/test-common -> ../../common/test-common
+e
 s 00053/00000/00000
 d D 1.1 10/04/29 02:05:14 joerg 1 0
 c date and time created 10/04/29 02:05:14 by joerg
@@ -15,7 +19,12 @@
 # format.sh:  Testing for correct interpretation of the format string.
 
 # Import common functions & definitions.
+D 2
 . ../common/test-common
+E 2
+I 2
+. ../../common/test-common
+E 2
 
 
 # If we invert the order of the arguments to prs here, so that the
